<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.2/1886137
 */

namespace Redis\Model;

use Application\Cache\AbstractCacheService;
use Application\Config\ConfigException;
use Application\Config\ConfigManager;
use Application\Connection\ConnectionFactory;
use Application\Lock\ILock;
use Application\Log\SwarmLogger;
use Application\Model\AbstractDAO as ApplicationDAO;
use Application\Model\IModelDAO;
use Closure;
use Interop\Container\ContainerInterface;
use P4\Connection\ConnectionInterface;
use P4\Log\Logger;
use P4\Model\Fielded\Iterator;
use P4\Spec\PluralAbstract;
use Record\Key\AbstractKey;
use Redis\Model\IModelDAO as RedisIModelDAO;
use Zend\ServiceManager\Exception\ServiceNotCreatedException;
use Exception;

/**
 * Abstract DAO for other redis cached based DAO implementations to use.
 * @package Redis\Model
 */
abstract class AbstractDAO extends ApplicationDAO implements RedisIModelDAO
{
    // This is the default PHP 'max_execution_time' that is applied when no value is set.
    // See: https://doc.bccnsoft.com/docs/php-docs-7-en/function.set-time-limit.html
    const DEFAULT_MAX_EXECUTION_TIME = 30;

    // Key and value used to put into cache.
    const POPULATED_STATUS = "populated-status";
    const POPULATED        = 1;
    const UNPOPULATED      = 0;
    const SCAN_PAGE_SIZE   = 1000;
    // When populating null means no maximum by default
    const FETCH_MAXIMUM = null;
    // Model code
    const MODEL = '';
    // The key for the verify status of the user dataset
    const VERIFY_STATUS = "verify-status";
    // Default maximum number of seconds to hold a lock
    const DEFAULT_POPULATION_LOCK_TIMEOUT = 300;
    // The template message for verification progress
    const VERIFY_PROGRESS = "Step %d of %d: %s";
    // The steps performed during verification
    const VERIFY_GET_REDIS_KEYS       = 'getRedisKeys';
    const VERIFY_GET_REDIS_CHECKSUMS  = 'getRedisChecksums';
    const VERIFY_GET_SERVER_CHECKSUMS = 'getServerChecksums';
    const VERIFY_REMOVE_EXTRA_KEYS    = 'removeExtraneousRedisKeys';
    const VERIFY_ADD_MISSING_KEYS     = 'addMissingModelKeys';
    // The progress message for all done
    const VERIFY_COMPLETE_MESSAGE = 'Verification complete';

    /**
     * AbstractDAO constructor. If the redis server does not have the populated status set we will rebuild the cache to
     * ensure it is up to date
     *
     * @param ContainerInterface $services
     * @param array|null         $options
     *
     * @throws ConfigException
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        parent::__construct($services, $options);
        // Get the server and check the current status or population.
        $this->populate();
    }

    /**
     * Check the the Id exists in the Cache first then fall back to the parent.
     *
     * @inheritDoc
     */
    public function exists($id, ConnectionInterface $connection = null)
    {
        $cacheService = $this->getCache();
        $multiple     = is_array($id);
        $exists       = false;
        $valid        = [];
        $connection   = $this->getConnection($connection);
        // Deal with arrays as well as single ids
        foreach ((array)$id as $id) {
            $id = $this->normalizeId($id, $connection);
            if ($cacheService) {
                $exists = $cacheService->has(static::CACHE_KEY_PREFIX . $id);
            }
            // Also check the Perforce server; cache may not be available or spec only exists in server
            if (!$exists) {
                $exists = parent::exists($id, $connection);
            }
            if ($exists) {
                $valid[] = $id;
            }
        }
        return $multiple ? $valid : $exists;
    }

    /**
     * Check the cache for the records and return if exists. Otherwise fall back to the parent.
     *
     * @inheritDoc
     */
    public function fetchById($id, ConnectionInterface $connection = null)
    {
        $connection   = $this->getConnection($connection);
        $cacheService = $this->getCache();
        $model        = null;
        // normalizeId for the User class will lowercase the id. This is correct
        // but we do not want to pass a lowercase id to parent::fetchById as in
        // PluralAbstract it will return a sparsely populated model with that
        // identifier set on it and we do not want the model to have the identifier
        // changed. 'exists' in model/User (which parent::fetchById calls) will
        // still work correctly as it uses lowercase id to query if the server
        // is case insensitive (Jira SW-6822)
        $nid = $this->normalizeId($id, $connection);
        if ($cacheService) {
            $model = $cacheService->get(static::CACHE_KEY_PREFIX . $nid);
            $model = $model ? $model : parent::fetchById($id, $connection);
        } else {
            $model = parent::fetchById($id, $connection);
        }
        return $model ? $this->setConnection($model, $connection) : null;
    }

    /**
     * Fetch all the records for the given model from the cache, otherwise fall back to Perforce fetchAll
     *
     * @inheritDoc
     */
    public function fetchAll(array $options = [], ConnectionInterface $connection = null)
    {
        // Get the fetch by constant from the model. Most use ids
        $fetchBy = constant(static::MODEL . '::FETCH_BY_IDS');
        $options = $options + [
            PluralAbstract::FETCH_MAXIMUM   => null,
            IModelDAO::FETCH_NO_CACHE       => null,
            $fetchBy                        => null
        ];
        if ($options[IModelDAO::FETCH_NO_CACHE]) {
            return parent::fetchAll($options, $connection);
        } else {
            $cacheService = $this->getCache();
            if ($cacheService) {
                $models = new Iterator();
                $max    = $options[PluralAbstract::FETCH_MAXIMUM];
                // If we are not fetching by ID/name we do the normal scan
                if (is_null($options[$fetchBy])) {
                    $keys = $this->getKeysFromCache($cacheService, $max);
                } else {
                    // If we are fetching by NAME/ID we should just use get multiple for them.
                    // This is faster than doing a scan though the full system.
                    $ids  = (array) $options[$fetchBy];
                    $keys = array_map(
                        function ($id) use ($connection) {
                            return static::CACHE_KEY_PREFIX.$this->normalizeId($id, $connection);
                        },
                        $ids
                    );
                }
                if (count($keys) > 0) {
                    foreach ($cacheService->getMultiple($keys) as $key => $model) {
                        if ($model) {
                            $models[$model->getId()] = $this->setConnection($model, $connection);
                        }
                    }
                }
                return $models;
            }
        }
        return parent::fetchAll($options, $connection);
    }

    /**
     * Ensure that the Redis cache data matches that in the Perforce server
     *
     * This is done in 5 steps:
     *   - 1) Getting the raw data from redis
     *   - 2) Getting the serialized version of the models
     *   - 3) Comparing the md5sums of the results
     *   - 4) Removing any keys which no longer have matching server data
     *   - 5) Caching any models which are not present
     *
     * Note that updated data may be deleted then added.
     */
    public function verify()
    {
        $this->logLine(Logger::DEBUG, __LINE__, "Verifying the Redis cache contents");
        $cache = $this->getCache();

        // Get checksums for the cache and the server
        $redisKeyChecksums      = $this->getRedisKeyChecksums($cache);
        $perforceModelChecksums = $this->getServerKeyChecksums();
        $this->logLine(
            Logger::TRACE,
            __LINE__,
            'Redis has (' . count($redisKeyChecksums) . ') keys.'.
            'Swarm has (' . count($perforceModelChecksums) . ') models, '
        );

        // Update the cache with any discrepancies
        $this->deleteExtraneousRedisKeys($cache, $redisKeyChecksums, $perforceModelChecksums);
        $this->addMissingRedisKeys($redisKeyChecksums, $perforceModelChecksums);

        // All done
        $this->setIntegrityStatus(static::VERIFY_COMPLETE_MESSAGE);
        $this->logLine(Logger::DEBUG, __LINE__, static::VERIFY_COMPLETE_MESSAGE);
    }

    /**
     * TODO: This is an example of the sort of thing we may want to expose just to indicate how it would operate.
     * TODO: It passes on the 'search' to a redis hGet call
     * @inheritDoc
     */
    public function search($key, $hashKey)
    {
        return $this->getCache()->getRedisResource()->hGet($key, $hashKey);
    }

    /**
     * Fetch the model by id and set it in the cache (without calling save on the model). Used to update the cache
     * when a new record has been added to perforce. The item will first be removed from the cache so that if it
     * has been deleted the fetch will not repopulate the cache.
     * @param string    $id     the id of the record, for example a user id. Cache key prefixes will be
     *                          used as per the implementing class
     * @param ConnectionInterface|null $connection the P4 connection
     * @return mixed the model
     */
    public function fetchByIdAndSet($id, ConnectionInterface $connection = null)
    {
        // normalizeId for the User class will lowercase the id. This is correct
        // but we do not want to pass a lowercase id to parent::fetchById as in
        // PluralAbstract it will return a sparsely populated model with that
        // identifier set on it and we do not want the model to have the identifier
        // changed. 'exists' in model/User (which parent::fetchById calls) will
        // still work correctly as it uses lowercase id to query if the server
        // is case insensitive (SW-6822)
        $nid          = $this->normalizeId($id);
        $model        = null;
        $cacheService = $this->getCache();
        $cacheService->delete(static::CACHE_KEY_PREFIX . $nid);
        $this->deleteFromSet($nid, $cacheService);
        try {
            $model = $this->fetchById($id, $connection);
            $cacheService->set(static::CACHE_KEY_PREFIX . $nid, $model);
            $this->addToSet([static::CACHE_KEY_PREFIX . $nid]);
        } catch (Exception $e) {
            $this->services->get(SwarmLogger::SERVICE)->debug("$nid removed from the cache, fetchById found no record");
        }
        return $model ? $this->setConnection($model, $connection) : null;
    }

    /**
     * @inheritDoc
     *
     * @return bool/null
     * @throws ConfigException
     */
    public function invalidate()
    {
        $result       = [];
        $cacheService = $this->getCache();

        $invalidate = function () use ($cacheService, &$result) {
            $this->deleteKeysForPrefix();
            $result[] = $cacheService->set(static::POPULATED_STATUS, self::UNPOPULATED);
        };

        // Ensure we don't run invalidate unless the cache is already populated
        $this->runUnderPopulateLock($cacheService, $invalidate, true);
        return empty($result) ? false : $result[0];
    }

    /**
     * @inheritDoc
     *
     * @return bool/null
     * @throws ConfigException
     */
    public function populate()
    {
        $result       = [];
        $cacheService = $this->getCache();

        // Callback to do the actual population
        $populate = function () use ($cacheService, &$result) {
            $this->deleteKeysForPrefix();
            $keyCount = 0;
            $lastSeen = null;
            // Fail back to the standard fetchAll function.
            do {
                $options = [
                    AbstractKey::FETCH_MAXIMUM => static::FETCH_MAXIMUM,
                    AbstractKey::FETCH_AFTER => $lastSeen
                ];
                $this->logger->debug(
                    "Redis/AbstractDAO::populate: fetchAll on "
                    . static::MODEL
                    . " with options [" . var_export($options, true) . "]"
                );
                $models = call_user_func(
                    static::MODEL . '::fetchAll',
                    $options,
                    $this->services->get(ConnectionFactory::P4_ADMIN)
                );
                // Now build the cache with all the model object to be cached.
                $modelKeys = $this->generateModelKeys($models, true);
                $keyCount += count($modelKeys);
                // Push cache items into the cache for the fetch
                $result[] = $cacheService->populateMultiple($modelKeys);
                $end      = end($models);
                $lastSeen = $end === false ? $end : $end->getId();
            } while (static::FETCH_MAXIMUM && $models && count($models) > 0);
            // setting the populated status key.
            $cacheService->set(static::POPULATED_STATUS, self::POPULATED);
            $this->logger->debug("Redis/AbstractDAO::populate: Saved $keyCount Redis keys for " . static::MODEL);
        };

        // Only populate if the cache is not already populated and it's not in the process of being populated
        $this->runUnderPopulateLock($cacheService, $populate, false);
        return empty($result) ? false : $result[0];
    }

    /**
     * Do the parent delete, then remove keys form redis.
     *
     * @inheritDoc
     */
    public function delete($model)
    {
        $result       = parent::delete($model);
        $cacheService = $this->getCache();
        if ($cacheService) {
            $cacheService->delete(static::CACHE_KEY_PREFIX . $model->getId());
            $this->deleteFromSet($model->getId(), $cacheService);
        }
        return $result;
    }

    /**
     * Get the value of the status.
     *
     * @inheritDoc
     */
    public function getIntegrityStatus()
    {
        $status       = false;
        $cacheService = $this->getCache();
        if ($cacheService->has(static::VERIFY_STATUS)) {
            $status = $cacheService->get(static::VERIFY_STATUS);
        }
        return $status;
    }

    /**
     * Set the value of the status.
     *
     * @inheritDoc
     */
    public function setIntegrityStatus($status = self::STATUS_QUEUED)
    {
        $cacheService = $this->getCache();
        return $cacheService->set(static::VERIFY_STATUS, $status);
    }

    /**
     * Delete the model key from the set
     * @param mixed $modelId        id of the model to delete
     * @param mixed $cacheService   the cache service
     */
    protected function deleteFromSet($modelId, $cacheService)
    {
        if ($cacheService) {
            $cacheService->getRedisResource()->sRemove(
                $this->getModelSetValue($cacheService),
                $cacheService->getNamespace() . static::CACHE_KEY_PREFIX . $modelId
            );
        }
    }

    /**
     * Generate the model keys. Then do parent save. Set the redis cache from the model keys.
     *
     * @inheritDoc
     */
    public function save($model)
    {
        $this->logger->info('Redis/AbstactDAO::save. Saving [' . $model->getId() . ']');
        $cacheService = $this->getCache();
        // If we don't have a ID we will save the model and then generate the model data.
        if ($model->getId() === null) {
            $model = parent::save($model);
            $keys  = $this->generateModelKeys([$model]);
        } else {
            $keys  = $this->generateModelKeys([$model]);
            $model = parent::save($model);
        }
        $this->logger->debug("Redis/AbstractDAO::save: Saving Redis keys " . implode(",", array_keys($keys)));
        $cacheService->setMultiple($keys);
        return $model;
    }

    /**
     * Delete all keys for the namespace and cache prefix defined for the DAO
     */
    private function deleteKeysForPrefix()
    {
        $cacheService    = $this->getCache();
        $namespace       = $cacheService->getNamespace();
        $namespaceLength = strlen($namespace);
        // Fetch all the keys that are prefix with this model and then delete them.
        $keysToDelete = array_map(
            function ($fullKey) use ($namespaceLength) {
                return substr($fullKey, $namespaceLength);
            },
            call_user_func_array(
                'array_merge',
                array_map(
                    function ($prefix) use ($cacheService, $namespace) {
                        return $cacheService->getRedisResource()->keys($namespace . $prefix . '*');
                    },
                    $this->getModelKeyPrefixes()
                )
            )
        );
        $cacheService->deleteMultiple($keysToDelete);
        $cacheService->getRedisResource()->del($this->getModelSetValue($cacheService));
    }

    /**
     * Get the cache service
     * @return mixed|null the cache service or null if it is not available
     */
    protected function getCache()
    {
        $cacheService = null;
        try {
            $cacheService = $this->services->get(AbstractCacheService::CACHE_SERVICE);
        } catch (ServiceNotCreatedException $snce) {
            // Ignore, cache not available
        }
        return $cacheService;
    }

    /** Get redis keys data from the cache, with an optional limit
     * @param $cache
     * @param int $max
     * @return array
     */
    protected function getKeysFromCache($cache, $max = 0)
    {
        $keys            = [];
        $redis           = $cache->getRedisResource();
        $namespace       = $cache->getNamespace();
        $namespaceLength = strlen($namespace);
        $scan            = $namespace.static::CACHE_KEY_PREFIX.'*';
        $current         = null;
        do {
            $nextKeys = $redis->sScan(
                $this->getModelSetValue($cache), $current, $scan, static::SCAN_PAGE_SIZE
            );
            $keys     = array_merge(
                $keys, array_map(
                    function ($fullKey) use ($namespaceLength) {
                        return substr($fullKey, $namespaceLength);
                    },
                    $nextKeys ?: []
                )
            );
            if ($max > 0 && count($keys) >= $max) {
                $keys = array_slice($keys, 0, $max);
                break;
            }
        } while ($nextKeys !== false);
        return $keys;
    }

    /**
     * Generate extra indices in the cache for the models (other than the standard id link). By default
     * we do not create any extra indices, specific implementations will handle their own
     *
     * @param mixed   $models      The Perforce models to generate indices for
     * @param boolean $rebuild     Build all available primary and any secondary model keys
     * @return array
     */
    protected function generateModelKeys($models, $rebuild = false)
    {
        $this->logger->info('Redis/AbstractDAO::generateModelKeys');
        $modelKeys = [];
        // Default implementation is to create no extra indices
        foreach ($models as $model) {
            $modelKeys[$this->buildModelKeyId($model)] = $model;
        }
        $this->addToSet(array_keys($modelKeys));
        return $modelKeys;
    }

    /**
     * Adds the keys to a redis set.
     * @param array         $modelKeys  keys to add
     * @param string|null   $setName    set name to use. Defaults to $this->getModelSetValue() if not provided
     */
    protected function addToSet(array $modelKeys, string $setName = null)
    {
        $cacheService = $this->getCache();
        $redis        = $cacheService->getRedisResource();
        $namespace    = $cacheService->getNamespace();
        $redis->sAddArray(
            $setName ? $setName : $this->getModelSetValue($cacheService),
            array_map(
                function ($value) use ($namespace) {
                    return $namespace . $value;
                },
                $modelKeys
            )
        );
    }

    protected function getModelKeyPrefixes()
    {
        return [static::CACHE_KEY_PREFIX];
    }

    /**
     * Build a redis key identifier from the cache prefix for this model and its ID.
     *
     * @param mixed   $model   This is the model we are dealing with.
     * @return string   This is the ID for this model.
     */
    protected function buildModelKeyId($model)
    {
        return static::CACHE_KEY_PREFIX .
            $this->normalizeId($model->getId(), $this->getConnection($model->getConnection()));
    }

    /**
     * @param object      $cacheService         cache service
     * @param Closure     $code                 code to run in the lock
     * @param bool        $shouldBePopulated    whether to check if populated or unpopulated
     *
     * @throws ConfigException
     */
    protected function runUnderPopulateLock($cacheService, $code, $shouldBePopulated)
    {
        // Get the name of the mutex and the timeout
        $mutexName = $this->getPopulateMutexName($cacheService);
        $timeout   = $this->getPopulateTimeout();

        // Callback to check if the cache is already populated
        $check = $this->checkPopulatedCallback($cacheService, $shouldBePopulated);

        // Run the code under the lock
        $lockService = $this->services->get(ILock::SERVICE);
        $this->extendMaxPhpExecutionTime($timeout);
        $lockService->lockWithCheck($mutexName, $check, $code, $timeout);
    }

    /**
     * Ensure that php code doesn't timeout before the lock is released
     *
     * @param int   $minimumSeconds    The number of seconds that the lock will be held
     */
    protected function extendMaxPhpExecutionTime($minimumSeconds)
    {
        $maxExecutionTime = ini_get('max_execution_time');

        // There is no limit on PHP code execution, so we have nothing to do
        if ($maxExecutionTime === '0') {
            return;
        }

        $maxExecutionTime = $maxExecutionTime === '' ? self::DEFAULT_MAX_EXECUTION_TIME : (int) $maxExecutionTime;

        // The code could timeout before the lock is released, so we extend the limit to match the lock timeout
        if ($maxExecutionTime < $minimumSeconds) {
            ini_set('max_execution_time', $minimumSeconds);
        }
    }

    /**
     * Constructs the check callback to check whether or not the cache is populated
     *
     * @param   mixed/null  $cacheService
     * @param   bool        $shouldBePopulated  Determines whether it's checking for populated or unpopulated state
     *
     * @return  Closure
     */
    protected function checkPopulatedCallback($cacheService, $shouldBePopulated)
    {
        $check = function () use ($cacheService, $shouldBePopulated): bool {
            $status = $cacheService->get(static::POPULATED_STATUS);
            return $shouldBePopulated ? $status == self::POPULATED : $status == self::UNPOPULATED;
        };

        return $check;
    }

    /**
     * Builds a unique name for the mutex
     *
     * @param $cacheService
     * @return string
     */
    protected function getPopulateMutexName($cacheService)
    {
        return $this->getModelSetValue($cacheService) . '_populate';
    }

    /**
     * Generate a name for a redis set based on the model
     * @param mixed     $cacheService       the cache service
     * @return string the model set name
     */
    protected function getModelSetValue($cacheService)
    {
        return $cacheService->getNamespace() . stripslashes(static::MODEL);
    }

    /**
     * Gets the population_timeout value from the config or returns a default value
     *
     * @return int
     */

    /**
     * @return array|bool|int|mixed|string|null
     * @throws ConfigException
     */
    protected function getPopulateTimeout()
    {
        $config = $this->services->get(ConfigManager::CONFIG);

        return ConfigManager::getValue(
            $config,
            ConfigManager::REDIS_POPULATION_LOCK_TIMEOUT,
            self::DEFAULT_POPULATION_LOCK_TIMEOUT
        );
    }

    /**
     * Builds an array of md5sum/Redis key mappings
     *
     * @param mixed|null    $cache      the cache service or null if it is not available
     *
     * @return array
     */
    protected function getRedisKeyChecksums($cache)
    {
        $namespace = $cache->getNamespace();

        $this->logLine(Logger::TRACE, __LINE__, "Getting all keys in the $namespace dataset.");
        $this->setIntegrityStatusStep(static::VERIFY_GET_REDIS_KEYS);
        $redisKeys = array_map(
            function ($key) use ($namespace) {
                return $namespace . $key;
            },
            $this->getKeysFromCache($cache)
        );

        $this->logLine(Logger::TRACE, __LINE__, 'Building md5sums for raw redis entries.');
        $this->setIntegrityStatusStep(static::VERIFY_GET_REDIS_CHECKSUMS);

        return $redisKeys ?
            array_flip(
                array_combine(
                    $redisKeys,
                    array_map(
                        function ($value) {
                            return md5($value);
                        },
                        $cache->getRedisResource()->mget($redisKeys)
                    )
                )
            )
            : [];
    }

    /**
     * Builds an array of md5sum/model id key mappings
     *
     * @return array|null
     */
    protected function getServerKeyChecksums()
    {
        $this->logLine(Logger::TRACE, __LINE__, 'Building md5sums for serialized models.');
        $this->setIntegrityStatusStep(static::VERIFY_GET_SERVER_CHECKSUMS);
        return array_flip(
            array_map(
                function ($value) {
                    return md5(serialize($value));
                },
                iterator_to_array(parent::fetchAll([]))
            )
        );
    }

    /**
     * Removes any keys found in Redis that are not also found in the Perforce server
     *
     * @param mixed|null    $cache              the cache service or null if it is not available
     * @param array         $redisEntries       md5sum/Redis key mappings
     * @param array         $perforceEntries    md5sum/model id key mappings
     */
    protected function deleteExtraneousRedisKeys($cache, $redisEntries, $perforceEntries)
    {
        $this->setIntegrityStatusStep(static::VERIFY_REMOVE_EXTRA_KEYS);

        $redis         = $cache->getRedisResource();
        $modelSetValue = $this->getModelSetValue($cache);

        foreach (array_diff_key($redisEntries, $perforceEntries) as $deleted) {
            $this->logLine(Logger::TRACE, __LINE__, "Deleting $deleted from dataset & redis");
            $redis->delete($deleted);
            $redis->sRemove($modelSetValue, $deleted);
        }
    }

    /**
     * Adds any keys found in Perforce that are not also found in Redis
     *
     * @param array    $redisEntries       md5sum/Redis key mappings
     * @param array    $perforceEntries    md5sum/model id key mappings
     */
    protected function addMissingRedisKeys($redisEntries, $perforceEntries)
    {
        // Any md5sums not in redis but in perforce need to be added
        $this->setIntegrityStatusStep(static::VERIFY_ADD_MISSING_KEYS);
        foreach (array_diff_key($perforceEntries, $redisEntries) as $unCached) {
            $this->logLine(Logger::TRACE, __LINE__, "Adding $unCached to dataset & redis");
            $this->fetchByIdAndSet($unCached);
        };
    }

    /**
     * Sets the verify-status key in Redis to a verification step with a description of the step
     *
     * @param string    $step    the name of the step being performed
     */
    protected function setIntegrityStatusStep($step)
    {
        $messages   = [
            static::VERIFY_GET_REDIS_KEYS       => 'Getting Redis cache entries',
            static::VERIFY_GET_REDIS_CHECKSUMS  => 'Calculating checksums for Redis keys',
            static::VERIFY_GET_SERVER_CHECKSUMS => 'Calculating checksums for Perforce models',
            static::VERIFY_REMOVE_EXTRA_KEYS    => 'Clearing any extraneous Redis cache entries',
            static::VERIFY_ADD_MISSING_KEYS     => 'Indexing any missing Perforce models into the Redis cache'
        ];
        $stepNumber = array_search($step, array_keys($messages))+1;
        $this->setIntegrityStatus(
            sprintf(static::VERIFY_PROGRESS, $stepNumber, count($messages), $messages[$step])
        );
    }

    /**
     * Logs a message, indicating the class and line number from where it is called
     *
     * @param int       $level      the log level to be applied
     * @param int       $line       the line number where this method is called
     * @param string    $message    the message to be logged
     */
    protected function logLine($level, $line, $message)
    {
        $this->logger->log($level, get_class($this) . ':' . __CLASS__ . " ($line): $message");
    }
}
