<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.2/1886137
 */

namespace Workflow\Model;

use Application\Model\ServicesModelTrait;
use Groups\Model\Group;
use Record\Exception\NotFoundException;
use Record\Key\AbstractKey;
use P4\Connection\ConnectionInterface as Connection;

/**
 * Model to define a workflow
 * @package Workflow\Model
 */
class Workflow extends AbstractKey implements IWorkflow
{
    const KEY_PREFIX    = 'swarm-workflow-';
    const UPGRADE_LEVEL = 6;
    const KEY_COUNT     = 'swarm-workflow:count';

    protected $fields = [
        IWorkflow::ON_SUBMIT => [
            'accessor' => 'getOnSubmit',
            'mutator'  => 'setOnSubmit',
            'index'    => 1702,
            'indexer'  => 'indexOnSubmit',
            'indexWords' => true
        ],
        IWorkflow::NAME => [
            'accessor' => 'getName',
            'mutator'  => 'setName',
            'index'    => 1701
        ],
        IWorkflow::DESCRIPTION => [
            'accessor' => 'getDescription',
            'mutator'  => 'setDescription'
        ],
        IWorkflow::SHARED => [
            'accessor'  => 'isShared',
            'mutator'   => 'setShared'
        ],
        IWorkflow::OWNERS => [
            'accessor' => 'getOwners',
            'mutator'  => 'setOwners'
        ],
        IWorkflow::END_RULES => [
            'accessor' => 'getEndRules',
            'mutator'  => 'setEndRules',
            'index'    => 1703,
            'indexer'  => 'indexEndRules',
            'indexWords' => true
        ],
        IWorkflow::AUTO_APPROVE => [
            'accessor' => 'getAutoApprove',
            'mutator'  => 'setAutoApprove',
            'index'    => 1704,
            'indexer'  => 'indexAutoApprove',
            'indexWords' => true
        ],
        IWorkflow::COUNTED_VOTES => [
            'accessor' => 'getCountedVotes',
            'mutator'  => 'setCountedVotes',
            'index'    => 1705,
            'indexer'  => 'indexCountedVotes',
            'indexWords' => true
        ],
        IWorkflow::GROUP_EXCLUSIONS => [
            'accessor' => 'getGroupExclusions',
            'mutator'  => 'setGroupExclusions',
            'index'    => 1706,
            'indexer'  => 'indexGroupExclusions',
            'indexWords' => true
        ],
        IWorkflow::USER_EXCLUSIONS => [
            'accessor' => 'getUserExclusions',
            'mutator'  => 'setUserExclusions',
            'index'    => 1707,
            'indexer'  => 'indexUserExclusions',
            'indexWords' => true
        ]
    ];

    protected function isCustomSearchExpression()
    {
        return true;
    }

    /**
     * Produces a 'p4 search' expression for the given field/value pairs.
     *
     * Extends parent to put the branches and project level workflow search
     * together.
     *
     * @param   array   $conditions     field/value pairs to search for
     *
     * @return  string  a query expression suitable for use with p4 search
     */
    protected static function makeSearchExpression($conditions)
    {
        return implode(
            ') | (',
            explode(
                ") (",
                str_replace(
                    '|||',
                    '|',
                    implode(
                        '|',
                        explode(
                            " ",
                            parent::makeSearchExpression($conditions)
                        )
                    )
                )
            )
        );
    }

    /**
     * @inheritDoc
     */
    public function getAutoApprove()
    {
        $autoApprove = $this->getRawValue(IWorkflow::AUTO_APPROVE);
        if (!is_array($autoApprove)) {
            $autoApprove = json_decode($autoApprove, true);
        }
        return $autoApprove !== null && $autoApprove !== false ? $autoApprove : [];
    }

    /**
     * @inheritDoc
     */
    public function setAutoApprove($autoApprove)
    {
        return $this->setRawValue(IWorkflow::AUTO_APPROVE, $autoApprove);
    }

    /**
     * @inheritDoc
     */
    public function setCountedVotes($countedVotes)
    {
        return $this->setRawValue(IWorkflow::COUNTED_VOTES, $countedVotes);
    }

    /**
     * @inheritDoc
     */
    public function getCountedVotes()
    {
        $countedVotes = $this->getRawValue(IWorkflow::COUNTED_VOTES);
        if (!is_array($countedVotes)) {
            $countedVotes = json_decode($countedVotes, true);
        }
        return $countedVotes !== null && $countedVotes !== false ? $countedVotes : [];
    }

    /**
     * @inheritDoc
     */
    public function setEndRules($endRules)
    {
        return $this->setRawValue(IWorkflow::END_RULES, $endRules);
    }

    /**
     * @inheritDoc
     */
    public function getEndRules()
    {
        $endRules = $this->getRawValue(IWorkflow::END_RULES);
        if (!is_array($endRules)) {
            $endRules = json_decode($endRules, true);
        }
        return $endRules !== null && $endRules !== false ? $endRules : [];
    }

    /**
     * @see IWorkflow::getOnSubmit()
     */
    public function getOnSubmit()
    {
        $onSubmit = $this->getRawValue(IWorkflow::ON_SUBMIT);
        if (!is_array($onSubmit)) {
            $onSubmit = json_decode($onSubmit, true);
        }
        return $onSubmit !== null && $onSubmit !== false ? $onSubmit : [];
    }

    /**
     * @see IWorkflow::setOnSubmit()
     */
    public function setOnSubmit($onSubmit)
    {
        return $this->setRawValue(IWorkflow::ON_SUBMIT, $onSubmit);
    }

    /**
     * @see IWorkflow::isOwner()
     */
    public function isOwner($userId)
    {
        $owned    = false;
        $groupDAO = ServicesModelTrait::getGroupDao();
        foreach ($this->getOwners() as $owner) {
            if (Group::isGroupName($owner)) {
                if ($groupDAO->isMember($userId, Group::getGroupName($owner), true, $this->getConnection())) {
                    $owned = true;
                    break;
                }
            } elseif ($owner === $userId) {
                $owned = true;
                break;
            }
        }
        return $owned;
    }

    /**
     * @see IWorkflow::canEdit()
     */
    public function canEdit($userIdentifier = null)
    {
        if ($userIdentifier instanceof Connection) {
            $editable = $this->isOwner($userIdentifier->getUser()) || $userIdentifier->isSuperUser();
        } elseif (is_string($userIdentifier)) {
            $editable = $this->isOwner($userIdentifier);
        } else {
            $editable = $this->isOwner($this->getConnection()->getUser());
        }
        return $editable;
    }

    /**
     * @see IWorkflow::getName()
     */
    public function getName()
    {
        return $this->getRawValue(IWorkflow::NAME);
    }

    /**
     * @see IWorkflow::setName()
     */
    public function setName($name)
    {
        return $this->setRawValue(IWorkflow::NAME, $name);
    }

    /**
     * @see IWorkflow::getDescription()
     */
    public function getDescription()
    {
        return $this->getRawValue(IWorkflow::DESCRIPTION);
    }

    /**
     * @see IWorkflow::setDescription()
     */
    public function setDescription($description)
    {
        return $this->setRawValue(IWorkflow::DESCRIPTION, $description);
    }

    /**
     * @see IWorkflow::isShared()
     */
    public function isShared()
    {
        return (bool) $this->getRawValue(IWorkflow::SHARED);
    }

    /**
     * @see IWorkflow::setShared()
     */
    public function setShared($shared)
    {
        return $this->setRawValue(IWorkflow::SHARED, (bool) $shared);
    }

    /**
     * @see IWorkflow::getOwners()
     */
    public function getOwners($flip = false)
    {
        return $this->getSortedField(IWorkflow::OWNERS, $flip);
    }

    /**
     * @see IWorkflow::setOwners()
     */
    public function setOwners($owners)
    {
        return $this->setRawValue(IWorkflow::OWNERS, $owners);
    }

    /**
     * @inheritDoc
     */
    public function getGroupExclusions()
    {
        $exclusions = $this->getRawValue(IWorkflow::GROUP_EXCLUSIONS);
        if (!is_array($exclusions)) {
            $exclusions = json_decode($exclusions, true);
        }
        return $exclusions !== null && $exclusions !== false ? $exclusions : [];
    }

    /**
     * @inheritDoc
     */
    public function setGroupExclusions($groupExclusions)
    {
        return $this->setRawValue(IWorkflow::GROUP_EXCLUSIONS, $groupExclusions);
    }

    /**
     * @inheritDoc
     */
    public function getUserExclusions()
    {
        $exclusions = $this->getRawValue(IWorkflow::USER_EXCLUSIONS);
        if (!is_array($exclusions)) {
            $exclusions = json_decode($exclusions, true);
        }
        return $exclusions !== null && $exclusions !== false ? $exclusions : [];
    }

    /**
     * @inheritDoc
     */
    public function setUserExclusions($userExclusions)
    {
        return $this->setRawValue(IWorkflow::USER_EXCLUSIONS, $userExclusions);
    }

    /**
     * Finds an item by its id.
     *
     * @param int|string $id the is to find
     * @param Connection $p4 the connection to use
     * @return Workflow|AbstractKey
     * @throws NotFoundException
     */
    public static function fetchById($id, Connection $p4 = null)
    {
        $p4 = $p4 ?: parent::getDefaultConnection();
        return parent::fetch($id, $p4);
    }

    /**
     * Gets the raw field value. Will sort alphabetically unless flip is requested.
     * @param $field
     * @param bool $flip
     * @return array|null
     */
    private function getSortedField($field, $flip = false)
    {
        $values = $flip
            ? array_flip((array) $this->getRawValue($field))
            : (array) $this->getRawValue($field);
        if (!$flip) {
            sort($values, SORT_STRING);
        }
        return $values;
    }

    /**
     * Extends parent to flip the ids ordering and hex encode.
     *
     * @param   string|int  $id     the user facing id
     * @return  string      the stored id used by p4 key
     */
    protected static function encodeId($id)
    {
        // nothing to do if the id is null
        if (!strlen($id)) {
            return null;
        }

        // subtract our id from max 32 bit int value to ensure proper sorting
        // we use a 32 bit value even on 64 bit systems to allow interoperability.
        $id = 0xFFFFFFFF - $id;

        // start with our prefix and follow up with hex encoded id
        // (the higher base makes it slightly shorter)
        $id = str_pad(dechex($id), 8, '0', STR_PAD_LEFT);
        return static::KEY_PREFIX . $id;
    }

    /**
     * Extends parent to undo our flip logic and hex decode.
     *
     * @param   string  $id     the stored id used by p4 key
     * @return  int             the user facing id
     */
    public static function decodeId($id)
    {
        // nothing to do if the id is null
        if ($id === null) {
            return null;
        }

        // strip off our key prefix
        $id = substr($id, strlen(static::KEY_PREFIX));

        // hex decode it and subtract from 32 bit int to undo our sorting trick
        return (int) (0xFFFFFFFF - hexdec($id));
    }

    /**
     * Get this workflow's values for the rules relating to submit commands.
     * @return array - the values of with and without review rules
     */
    public function indexOnSubmit()
    {
        $keywords = [];
        $onSubmit = $this->getOnSubmit();
        if (isset($onSubmit[IWorkflow::WITH_REVIEW][IWorkflow::RULE])) {
            $keywords[] = $onSubmit[IWorkflow::WITH_REVIEW][IWorkflow::RULE];
        }
        if (isset($onSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE])) {
            $keywords[] = $onSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE];
        }
        return $keywords;
    }

    /**
     * Get this workflow's value for the rules relating to 'ended' reviews.
     * @return string
     */
    public function indexEndRules()
    {
        $keyword  = '';
        $endRules = $this->getEndRules();
        if (isset($endRules[IWorkflow::UPDATE][IWorkflow::RULE])) {
            $keyword = $endRules[IWorkflow::UPDATE][IWorkflow::RULE];
        }
        return $keyword;
    }

    /**
     * Get this workflow's value for the rules related to reviews with all necessary votes.
     * @return string
     */
    public function indexAutoApprove()
    {
        return $this->indexField($this->getAutoApprove());
    }

    /**
     * Get this workflow's value for whose votes count towards requirements.
     * @return string
     */
    public function indexCountedVotes()
    {
        return $this->indexField($this->getCountedVotes());
    }

    /**
     * Get this workflow's value for user exclusions.
     * @return string
     */
    public function indexUserExclusions()
    {
        return $this->indexField($this->getUserExclusions());
    }

    /**
     * Get this workflow's value for group exclusions.
     * @return string
     */
    public function indexGroupExclusions()
    {
        return $this->indexField($this->getGroupExclusions());
    }

    /**
     * Index a simple field that specifies rule
     * @param array     $field  the field
     * @return string the index keyword
     */
    private function indexField($field)
    {
        $keyword = '';
        if (isset($field[IWorkflow::RULE])) {
            $keyword = $field[IWorkflow::RULE];
        }
        return $keyword;
    }
}
